"""Module for reading gshhs binary coastline formats"""

import struct, os
import numarray as NA
from matplotlib.mlab import conv

class gshhs:
    def _get_header(self):
        nb_i = struct.calcsize('i') # 4 bytes in 'int'
        nb_h = struct.calcsize('h') # 2 bytes in 'short int'
        hdr_len = nb_i*8 + nb_h*2
        bhdr = self.f.read(hdr_len)
        if not bhdr:
            return None
        hdr = {}; n=0
        # integer header information (big-endien)
        hdr['id']     = struct.unpack('>i',bhdr[n:n+nb_i])[0];        n += nb_i
        hdr['n']      = struct.unpack('>i',bhdr[n:n+nb_i])[0];        n += nb_i
        hdr['level']  = struct.unpack('>i',bhdr[n:n+nb_i])[0];        n += nb_i
        hdr['west']   = struct.unpack('>i',bhdr[n:n+nb_i])[0]*1.0e-6; n += nb_i
        hdr['east']   = struct.unpack('>i',bhdr[n:n+nb_i])[0]*1.0e-6; n += nb_i
        hdr['south']  = struct.unpack('>i',bhdr[n:n+nb_i])[0]*1.0e-6; n += nb_i
        hdr['north']  = struct.unpack('>i',bhdr[n:n+nb_i])[0]*1.0e-6; n += nb_i
        hdr['area']   = struct.unpack('>i',bhdr[n:n+nb_i])[0];        n += nb_i
        # short integer header information
        hdr['grenwich']  = struct.unpack('>h',bhdr[n:n+nb_h])[0];     n += nb_h
        hdr['source']    = struct.unpack('>h',bhdr[n:n+nb_h])[0];
        return hdr
    
    def _get_poly(self,n):
        n_i4 = struct.calcsize('i')
        bdata = self.f.read(2*n_i4*n)
	# unpack big-endien micro degrees
        data = NA.array(struct.unpack('>'+str(2*n)+'i',bdata)[:])
        lon = NA.array(data[::2]*1.0e-6)
        lat = NA.array(data[1::2]*1.0e-6)
        # Clip data to window
	lon = NA.clip(lon,self.window[0]%360,self.window[1]%360)
	lat = NA.clip(lat,self.window[2],self.window[3])
        # shrink corners to a single point
        b = NA.array([0.5, 0.5])
        idx = ( ((lon==self.window[0]%360)&(lat==self.window[2]))  # South-west corner
              | ((lon==self.window[1]%360)&(lat==self.window[2]))  # South-east corner
              | ((lon==self.window[0]%360)&(lat==self.window[3]))  # North-west corner
              | ((lon==self.window[1]%360)&(lat==self.window[3])) )# North-east corner
        idx=conv(idx,b,1)
        lon=lon[idx<1]; lat=lat[idx<1]
        # Only keep first and last points in series of identical values
 	b = NA.array([0.25, 0.5, 0.25])
        idx = ( (lon==self.window[0]%360)
              | (lon==self.window[1]%360)
              | (lat==self.window[2])
              | (lon==self.window[3]) )
 	idx = conv(idx,b,1)
        # Keep corner points
        idx[((lon==self.window[0]%360)&(lat==self.window[2]))
            | ((lon==self.window[1]%360)&(lat==self.window[2]))
            | ((lon==self.window[0]%360)&(lat==self.window[3]))
            | ((lon==self.window[1]%360)&(lat==self.window[3]))] = 0
 	lon=lon[idx<1]; lat=lat[idx<1]
        if len(lon) != len(lat):
            raise ValueError
	return lon.tolist(),lat.tolist()
           
    def __init__(self,resolution='i',window=None,area_clip=0.,data_dir=None):
        if data_dir is None:
            data_dir = os.environ.get('GSHHS_PATH')
        if window is None:
            window=[0.0001, 359.9999, -90., 90.]
        self.window = window
        self.f = open(os.path.join(data_dir,'gshhs_' + resolution + '.b'),'rb')
        self.coastlons=[]; self.coastlats=[];
        self.coastsegind=[]; self.coastsegtype=[]; self.coastsegarea=[]
        hdr = gshhs._get_header(self)
        while hdr:
            if (hdr['east']%360>window[0]%360 and hdr['west']%360<window[1]%360 and 
                hdr['north']>window[2] and hdr['south']<window[3] and 
                hdr['area']>area_clip):
                lons, lats = self._get_poly(hdr['n'])
                self.coastsegind.append(len(self.coastlons))
                self.coastsegtype.append(hdr['level'])
                self.coastsegarea.append(hdr['area'])
                [self.coastlons.append(lon) for lon in lons]
                [self.coastlats.append(lat) for lat in lats]
            else:
                self.f.seek(2*4*hdr['n'],1)
            hdr = gshhs._get_header(self)
	self.f.close()
	    
if __name__ == '__main__':
    import time,sys
    resolution = 'l'
    resolution = sys.argv[1]
    area_thresh = float(sys.argv[2])
    window=[-95., -87., 28., 31.0]
    t1 = time.clock()
    m = gshhs(window=window,resolution=resolution,area_clip=area_thresh)
    t2 = time.clock()
    print 'gshhs time (gulf coast subset): ' + str(t2-t1) + ' seconds'
    print len(m.coastlons)
    t1 = time.clock()
    m = gshhs(resolution=resolution,area_clip=area_thresh)
    t2 = time.clock()
    print 'gshhs time (full dataset): ' + str(t2-t1) + ' seconds'
    print len(m.coastlons)
